{-# LANGUAGE FlexibleInstances,
             MultiParamTypeClasses,
             UndecidableInstances,
             ScopedTypeVariables,
             FlexibleContexts,
             RankNTypes
 #-}

module AOP.Internal.AOT_secure (
 AOT_s,
 runAOT_s,
 wappt_s,
) where

import AOP.Internal.Typeable1Monad
import AOP.Internal.Aspect
import AOP.Internal.Pointcut
import AOP.Internal.Advice
import AOP.Internal.Joinpoint
import AOP.Internal.AOPMonad
import AOP.Internal.OpenApp
import AOP.Internal.Function
import AOP.Internal.AOTType


-- | Runs an AOT_s computation to obtain a computation in the underlying monad
runAOT_s :: Monad m => EPC (AOT_s m) -> AOT_s m a -> m a
runAOT_s pc c = do (a, _) <- run_s c [] pc
                   return a

-- | Monadic weaver
weavet_s :: Typeable1Monad m => 
            (a -> AOT_s m b) -> AspectEnv (AOT_s m) -> EPC (AOT_s m) -> Jp (AOT_s m) a b -> m (a -> AOT_s m b)
weavet_s f aenv pcG@(EPC _pcG) jp = do (match, _) <- run_s (runPC _pcG jp) aenv pcG
                                       if match
                                          then return f  
                                          else weavet_saux f aenv pcG jp

weavet_saux f [] pcG _ = return f
weavet_saux f aenv@(asp:asps) pcG jp = 
 case asp of EAspect (Aspect pc adv _) -> do (match, _) <- run_s (runPC pc jp) aenv pcG
                                             weavet_s (if match  
                                                          then (apply_adv adv f)
                                                          else f) asps pcG jp

-- | Implementation of woven application for AOT_s, used in the overloading of #.
-- FunctionTag argument is used for function identity:
--     Wrapped functions are equal when their tags are equal
--     Regular functions share the same tag, and are compared using StableNames.
wappt_s :: (Typeable1Monad m, PolyTypeable (a -> AOT_s m b)) => 
           (a -> AOT_s m b) -> FunctionTag -> a -> AOT_s m b
wappt_s f t a = AOT_s $ \aenv pc -> do woven_f <- weavet_s f aenv pc (newjp f t a)
                                       run_s (woven_f a) aenv pc

-- | Every regular functions is tagged with the same default tag.
instance Typeable1Monad m => OpenApp (->) (AOT_s m) where
         f # a = wappt_s f defaultFunctionTag a

-- | Function is a wrapper to add a notion of identity based on tags.
instance Typeable1Monad m => OpenApp Function (AOT_s m) where
         (Function f t) # a = wappt_s f t a

-- | Monadic bind and return operations for a transformed monad.
-- The aspect environment is wrapped or passed around as in a state monad transformer.
instance Monad m => Monad (AOT_s m) where
    return a = AOT_s $ \aenv pc -> return (a, (aenv,pc))
    k >>= f  = AOT_s $ \aenv pc -> do (arg, (aenv',pc')) <- run_s k aenv pc
                                      run_s (f arg) aenv' pc'

-- | Typeable instance so types of computations in AOT_s can be compared (like in pcCall and pcType)
instance Typeable1Monad m => Typeable1 (AOT_s m) where
         typeOf1 _ = mkTyConApp (mkTyCon3 "AOT_s" "AOT_s" "AOT_s") [typeOf1 (undefined :: m ())]

-- | The semantics of aspect deployment are defined in the
-- MonadDeploy typeclass. AOT_s assumes it is on top of an MonadDeploy
-- instance, and uses that functions for aspect deployment.
instance (Typeable1Monad m, MonadDeploy AOT_s m) => AOPMonad (AOT_s m) where
      deploy asp   = AOT_s $ \aenv pc -> do aenv' <- deployInEnv asp aenv
                                            return ((), (aenv',pc))
      undeploy asp = AOT_s $ \aenv pc -> do aenv' <- undeployInEnv asp aenv
                                            return ((), (deleteAsp (EAspect asp) aenv, pc))


-- Interaction with standard monad transformers

instance MonadTrans AOT_s where
         lift ma = AOT_s $ \aenv pc -> do
                           a <- ma
                           return (a, (aenv,pc))

instance (MonadIO m) => MonadIO (AOT_s m) where
         liftIO = lift . liftIO

instance (MonadState s m) => MonadState s (AOT_s m) where
         get = lift get
         put = lift . put

instance (MonadError s m) => MonadError s (AOT_s m) where
         throwError = lift . throwError
         -- ma `catchError` h = AOT_s $ \ctx -> do
         --        (a,ctx') <- run ma ctx
         --        case a of
         --             (Left  l) -> run (h l) ctx'
         --             (Right r) -> run ma ctx' -- return (Right r, ctx')

instance (MonadWriter w m) => MonadWriter w (AOT_s m) where
    tell     = lift . tell
    listen m = AOT_s $ \aenv pc -> do
                              ((a, ctx), w) <- listen (run_s m aenv pc)
                              return ((a, w), ctx)

    pass m = AOT_s $ \aenv pc -> pass $ do
                                   ((a, f), ctx) <- run_s m aenv pc
                                   return ((a, ctx), f)

-- instance MonadCont m => MonadCont (AOT_s m) where
--     callCC f = AOT_s $ \aenv -> \pc -> 
--         callCC $ \c ->
--         run_s (f (\a -> AOT_s $ \aenv' -> \pc' -> c (a, (aenv',pc')))) aenv pc

instance (MonadReader r m) => MonadReader r (AOT_s m) where
    ask       = lift ask
    local f m = AOT_s $ \aenv pc -> local f (run_s m aenv pc)



